import org.freedesktop.Avahi.*;
import org.freedesktop.dbus.*;
import java.util.*;
import java.net.InetAddress;

import org.snmp4j.*;
import org.snmp4j.smi.*;
import org.snmp4j.mp.*;
import org.snmp4j.event.*;
import org.snmp4j.transport.*;

public class Publish implements CommandResponder {
	final DBusConnection bus;
	final Server avahi;
	EntryGroup group;
	final Snmp snmp;
	
	List<DeviceProperty> publishProp = new ArrayList<DeviceProperty>();
	List<DeviceProperty> snmpProp = new ArrayList<DeviceProperty>();
	//SortedMap<OID,Variable> variables = new TreeMap<OID,Variable>();
	static final OID propPrefix = new OID(new int[]{1,0});
	int port;
	String name;

	public Publish() {
		port = new Random().nextInt(64000) + 1024;
		try {
		bus = DBusConnection.getConnection(DBusConnection.SYSTEM);
		avahi = (Server)bus.getRemoteObject(
			"org.freedesktop.Avahi", "/", Server.class);

		// SNMP testing support
		snmp = new Snmp(new DefaultUdpTransportMapping(new UdpAddress(port)));
		snmp.addCommandResponder(this);
		snmp.listen();
		} catch (java.io.IOException ex) {
			throw new RuntimeException(ex);
		} catch (org.freedesktop.dbus.exceptions.DBusException ex) {
			throw new RuntimeException(ex);
		}
	}
	protected List<Byte> stringToByte(String str) {
		try {
		byte[] field = str.getBytes("utf-8");
		Byte[] convField = new Byte[field.length];
		for (int x = 0; x < field.length; x++)
			convField[x] = field[x];
		return Arrays.asList(convField);
		} catch (java.io.UnsupportedEncodingException ex) {
			throw new RuntimeException(ex);
		}
	}
	
	public void addPublishProperty(DeviceProperty property) {
		publishProp.add(property);
	}
	public void addSnmpProperty(DeviceProperty property) {
		snmpProp.add(property);
	}
	public void publish(String name) {
		this.name = name;
		group = (EntryGroup)avahi.EntryGroupNew();
		List<List<Byte>> txt = new ArrayList<List<Byte>>();
		
		for (DeviceProperty prop: publishProp)
			txt.add(stringToByte(prop.toString()));
		group.AddService(-1, -1, new UInt32(0),
			name, "_hacs._udp", "",
			"", new UInt16(port), txt);
		group.Commit();
	}
	
	public synchronized void processPdu(CommandResponderEvent e) {
		PDU command = e.getPDU();
		if (command == null)
			return;
		try {
		
		PDU ret = new PDU(command);
		ret.setType(PDU.RESPONSE);
		if (command.getType() == PDU.GET) {
			for (VariableBinding vb: ret.toArray()) {
				OID curoid = vb.getOid();
				if (!curoid.startsWith(propPrefix))
					vb.setVariable(Null.noSuchObject);
				else if (curoid.size() != 4 || curoid.get(3) >= snmpProp.size() || curoid.get(2) > 1)
					vb.setVariable(Null.noSuchInstance);
				else {
					DeviceProperty prop = snmpProp.get(curoid.get(3));
					String value = curoid.get(2) == 0 ? prop.getName() : prop.getValue();
					vb.setVariable(new OctetString(value));
				}
			}
		} else if (command.getType() == PDU.GETNEXT) {
			for (VariableBinding vb: ret.toArray()) {
				OID curoid = vb.getOid().successor();
				if (!curoid.startsWith(propPrefix))
					vb.setVariable(Null.noSuchObject);
				else if (curoid.size() != 4 || curoid.get(2) > 1 || curoid.get(3) >= snmpProp.size() + 1)
					vb.setVariable(Null.noSuchInstance);
				else if (curoid.get(2) == 1 && curoid.get(3) >= snmpProp.size())
					vb.setVariable(Null.noSuchInstance);
				else {
					DeviceProperty prop = snmpProp.get(curoid.get(3));
					String value = curoid.get(2) == 0 ? prop.getName() : prop.getValue();
					vb.setVariable(new OctetString(value));
					vb.setOid(curoid);
				}
			}
		} else if (command.getType() == PDU.SET) {
			VariableBinding[] vbArray = ret.toArray();
			// Check for errors
			// TODO: set error indexes
			for (VariableBinding vb: vbArray) {
				if (vb.getOid().startsWith(new OID(propPrefix.getValue(),
						new int[]{1})))
					continue;
				ret.setErrorStatus(PDU.notWritable);
				sendResponse(e, ret);
				return;
			}
			for (VariableBinding vb: vbArray) {
				if (vb.getVariable() instanceof OctetString)
					continue;
				ret.setErrorStatus(PDU.wrongType);
				sendResponse(e, ret);
				return;
			}
			for (VariableBinding vb: vbArray) {
				if (vb.getOid().size() == 4 && vb.getOid().get(3) < snmpProp.size() && vb.getOid().get(2) <= 1)
					continue;
				ret.setErrorStatus(PDU.noCreation);
				sendResponse(e, ret);
				return;
			}
			// No errors, set fields
			for (VariableBinding vb: vbArray) {
				DeviceProperty prop = snmpProp.get(vb.getOid().get(3));
				prop.setValue(((OctetString)vb.getVariable()).toString());
			}
		} else {
			System.out.println("Unknown PDU; ignoring");
			return;
		}
		sendResponse(e, ret);
		
		} catch(Exception ex) { ex.printStackTrace(); }
	}
	private void sendResponse(CommandResponderEvent e, PDU response)
			throws MessageException {
		MessageDispatcher md = e.getMessageDispatcher();
		md.returnResponsePdu(e.getMessageProcessingModel(), e.getSecurityModel(),
			e.getSecurityName(), e.getSecurityLevel(), response,
			e.getMaxSizeResponsePDU(), e.getStateReference(), null);
		e.setProcessed(true);
	}
	public void disconnect() {
		try {
			snmp.close();
		} catch (java.io.IOException ex) {
			System.out.println("snmp.close() failed");
		}
		group.Free();
		bus.disconnect();
	}
	public static void main (String [] args) {
		Publish publish = new Publish();
		publish.addPublishProperty(new DeviceProperty("nothing", "here"));
		publish.addPublishProperty(new DeviceProperty("category", "media"));
		publish.addPublishProperty(new DeviceProperty("room", "default"));
		publish.publish("Sample HACS Device " + new Random().nextInt(100));
		publish.addSnmpProperty(new DeviceProperty("marquee", "hello world"));
		publish.addSnmpProperty(new DeviceProperty("hello", "world"));
		System.out.println("Press enter to stop...");
		try {
			System.in.read();
		} catch (java.io.IOException ex) { throw new RuntimeException(ex); }
		publish.disconnect();
	}
}
